from qgis.PyQt.QtCore import (
  QCoreApplication
  )
from qgis.core import (
  QgsVectorLayer,
  QgsProcessingAlgorithm,
  QgsProcessingParameterDefinition, 
  QgsProcessingParameterNumber,
  QgsProcessingUtils,
  QgsVectorFileWriter,
  QgsCoordinateTransformContext,
  QgsFeatureRequest,
  QgsProcessingParameterString,
  QgsProcessingParameterBoolean,
  QgsCoordinateReferenceSystem,
  QgsRasterFileWriter,
  QgsProcessingFeedback,
  QgsProcessingContext,
  QgsRasterLayer,
  QgsReferencedRectangle,
  QgsCoordinateTransform,
  QgsProject,
  QgsProcessing,
  Qgis
  )

import os
import asyncio
import re
import platform
import subprocess
import uuid

from qgis.utils import iface

class algabstract(QgsProcessingAlgorithm):
  
  PARAMETER = {}
  ENV = {}
  NOISEMODELLING = {}

  # initialization  
  def __init__(self):
    super().__init__()
    self.NOISEMODELLING = {
      "CMD": None,
      "GROOVY_SCRIPT": None,
      "TEMP_DIR":None,
      "WPS_ARGS":None
      }
        
    self.ENV = {
      "PLATFORM": platform.system()
    }
    
    try:
      # set environment paths
      self.ENV["NOISEMODELLING_HOME"]     = os.environ["NOISEMODELLING_HOME"]
      self.ENV["JAVA_FOR_NOISEMODELLING"] = os.environ["JAVA_FOR_NOISEMODELLING"]
    except:
      raise Exception("Error(s) occurred at setting Environment paths.")
        
  # set environment variables
  def setEnvVariable(self, key: str, value: str) -> None:
    try:
      if self.ENV["PLATFORM"] == 'Windows':          
        subprocess.call(["setx", key, value])
      else:
        subprocess.run(f'echo \'export {key}="{value}"\' >> ~/.bash_profile', shell=True, check=True)
    except:
      raise Exception("Error(s) occurred at setting Environment variables.")
  
  # add parameters using PARAMETERS attribute
  def initParameters(self) -> None:    
    for key, value in self.PARAMETERS.items():
      try:
        args = value.get("ui_args")
        args["name"] = key
        args["description"] = self.tr(args["description"])
                
        ui = value.get("ui_func")(**args)
        
        if value.get("advanced") is not None and value.get("advanced") == True:
          ui.setFlags(QgsProcessingParameterDefinition.FlagAdvanced)
          
        self.addParameter(ui)  
      except Exception as e:       
        self.iface.messageBar().pushMessage(str(e), self.tr("Error(s) occurred at creating a user interface of" + self.__class__ + "."), level=Qgis.Critical)
        raise Exception(e)
  
  def checkCrsAsCartesian(self, crs: QgsCoordinateReferenceSystem, feedback:QgsProcessingFeedback) -> None:
    if crs.isGeographic() or crs.authid() == "EPSG:3857":
      feedback.reportError(self.tr("The Target CRS is NOT a Cartesian Coordinate System."))
      raise Exception(self.tr("The Target CRS is NOT a Cartesian Coordinate System."))
  
  # get the fetch extent and crs using the canvas
  def getExtentAndCrsUsingCanvas(self, get_xy_coords= True) -> (QgsReferencedRectangle, QgsCoordinateReferenceSystem):
    try:
      map_rectangle = iface.mapCanvas().extent()
      proj_crs = iface.mapCanvas().mapSettings().destinationCrs()
      
      use_rectangle = QgsReferencedRectangle(map_rectangle, proj_crs)
      
      use_crs = proj_crs
      if get_xy_coords:
        # if the CRS is geographic, input the long/lat to the getUtmCrs
        if proj_crs.isGeographic():
          use_crs = self.getUtmCrs(map_rectangle.center().x(), map_rectangle.center().y())
        # if the CRS is web-mercator, transform long/lat coordinates and input to the getUtmCrs
        elif proj_crs.authid() == "EPSG:3857":
          transform = QgsCoordinateTransform(proj_crs, QgsCoordinateReferenceSystem("EPSG:4326"), QgsProject.instance())
          map_rectangle = transform.transformBoundingBox(map_rectangle)
          use_crs = self.getUtmCrs(map_rectangle.center().x(), map_rectangle.center().y())
            
      return use_rectangle, use_crs
    except:
      return None, None
  
      
  
  # get CRS of the UTM using longitude and latitude
  def getUtmCrs(self, lng: float, lat: float) -> QgsCoordinateReferenceSystem:
    epsg_code = 32600 + 100 * (lat < 0) + int((lng + 180) / 6) + 1
    crs = QgsCoordinateReferenceSystem(f'EPSG:{epsg_code}')
    if crs.isValid():
        return crs
    else:
        return None
  

  # download a file from the URL
  def downloadFile(self, session, request_args:dict, download_msg:str, feedback_progress: bool, feedback: QgsProcessingFeedback) -> str:
    response = session.request(stream = True, timeout = 12.0, **request_args)
    
    # if download is successful
    if response.status_code == 200:
      if len(download_msg) > 0:
        feedback.pushInfo(download_msg)
      file_size = int(response.headers.get('content-length', 0))
      
      file_path = os.path.join(
        os.path.normpath(os.path.dirname(QgsProcessingUtils.generateTempFilename(""))), 
        str(uuid.uuid4())
      )

      # write downloaded file and show progress
      with open(file_path, 'wb') as file:
        for data in response.iter_content(chunk_size=1024):
          file.write(data)
          if feedback_progress:
            feedback.setProgress(int((file.tell() / file_size) * 100))
      
      if feedback_progress:
        feedback.setProgress(0)
      
      return file_path
      
    else:
      # if download is failed
      raise Exception(self.tr("Download was failed."))
      
  # initialize NoiseModelling paths
  def initNoiseModellingPath(self, paths:dict, feedback:QgsProcessingFeedback) -> None:
    if paths.get("GROOVY_SCRIPT") is None:
      feedback.reportError(self.tr("Groovy script is not specified"), fatalError=True)
      raise Exception(self.tr("Groovy script is not specified"))
  
    # folder where the files are saved
    self.NOISEMODELLING["TEMP_DIR"] = os.path.normpath(os.path.dirname(QgsProcessingUtils.generateTempFilename("")))
    if not os.path.exists(self.NOISEMODELLING["TEMP_DIR"]):
      os.makedirs(self.NOISEMODELLING["TEMP_DIR"])
    
    if paths is not None and isinstance(paths, dict):
      for key, value in paths.items():
        if isinstance(value, str):
          paths[key] = value.replace("%nmtmp%", self.NOISEMODELLING["TEMP_DIR"])
        elif isinstance(value, dict):
          for key2, value2 in value.items():
            paths[key][key2] = value2.replace("%nmtmp%", self.NOISEMODELLING["TEMP_DIR"])
      self.NOISEMODELLING.update(paths)

  
  # initialize NoiseModelling arguments using UIs
  def initNoiseModellingArg(
    self, parameters:dict, context: QgsProcessingContext, feedback:QgsProcessingFeedback
    ) -> None:   
    
    # set arguments for wps_scripts
    # note that the paths are surrounded by double quotes
    self.NOISEMODELLING["WPS_ARGS"] = {
      "w": '"' + self.NOISEMODELLING["TEMP_DIR"] + '"', 
      "s": '"' + self.NOISEMODELLING["GROOVY_SCRIPT"] + '"',
      "noiseModellingHome": '"' + os.path.normpath(self.ENV["NOISEMODELLING_HOME"]) + '"',
      "exportDir": '"' + self.NOISEMODELLING["TEMP_DIR"]+ '"'
    }
    
    # get the reference CRS
    try:
      crs_key = [key for key, value in self.PARAMETERS.items() if value.get("crs_referrence") is not None and value.get("crs_referrence") == True]
      crs_referrence = self.parameterAsSource(parameters, crs_key[0], context).sourceCrs()
      self.NOISEMODELLING["WPS_ARGS"]["inputSRID"] = crs_referrence.authid().replace("EPSG:","")
    except:
      feedback.pushWarning("CRS is not specified!")
      crs_referrence = None
    
    # set other arguments
    for key, value in self.PARAMETERS.items():
      if value.get("n_mdl") is not None:
        if value.get("save_layer_get_path") is not None and value.get("save_layer_get_path") == True:
          src = self.parameterAsSource(parameters, key, context)
          if src is not None:
            if src.sourceCrs() != crs_referrence:
              feedback.reportError(self.tr("CRS is not the same among input features."), fatalError=True)
              raise Exception(self.tr("CRS is not the same among input features."))
            vl = src.materialize(QgsFeatureRequest(), feedback)
            vl_path = os.path.join(self.NOISEMODELLING["TEMP_DIR"], key + ".geojson")
            self.saveVectorLayer(vl, vl_path)
            # register in the WPS_ARGS
            self.NOISEMODELLING["WPS_ARGS"][value.get("n_mdl")] = '"' + vl_path + '"'
        else:
          if value.get("ui_func") == QgsProcessingParameterString:
            value_input = '"' + self.parameterAsString(parameters, key, context) + '"'
          if value.get("ui_func") == QgsProcessingParameterBoolean:
            value_input = self.parameterAsInt(parameters, key, context)
          if value.get("ui_func") == QgsProcessingParameterNumber:
            if value.get("ui_args").get("type") == QgsProcessingParameterNumber.Integer:
              value_input = self.parameterAsInt(parameters, key, context)
            else:
              value_input = self.parameterAsDouble(parameters, key, context)
            
          # register in the WPS_ARGS
          self.NOISEMODELLING["WPS_ARGS"][value.get("n_mdl")] = value_input
  
  # add NoiseModelling arguments manually
  def addNoiseModellingArg(self, args:dict=None) -> None:
    if (isinstance(args, dict)):
      self.NOISEMODELLING["WPS_ARGS"].update(args)
  
  # to save a vector layer
  def saveVectorLayer(self, vector_layer: QgsVectorLayer, path: str) -> None:
    save_options = QgsVectorFileWriter.SaveVectorOptions()
    save_options.driverName = "GeoJSON"
    QgsVectorFileWriter.writeAsVectorFormatV3(
      vector_layer, path, QgsCoordinateTransformContext(), save_options
    )
  
  
  # to save a raster layer
  def saveRasterLayer(self, raster_layer:QgsRasterLayer, path: str) -> None:
    file_writer = QgsRasterFileWriter(path)
    file_writer.writeRaster(
      raster_layer.pipe(), 
      raster_layer.width(), raster_layer.height(),
      raster_layer.dataProvider().extent(),
      raster_layer.crs()
      )
  
  # execution of the NoiseModelling script
  def execNoiseModellingCmd(
    self, parameters: dict, context: QgsProcessingContext, feedback: QgsProcessingFeedback
    ) -> None:    
    
    self.NOISEMODELLING["CMD"] = '"' + os.path.join(os.path.dirname(__file__), "noisemodelling","bin","wps_scripts") + '"' + \
      "".join([" -" + k + " " + str(v) for k, v in self.NOISEMODELLING["WPS_ARGS"].items()])
    feedback.pushCommandInfo("CMD FOR WPS_SCRIPTS: " + self.NOISEMODELLING["CMD"])   
    
    loop = asyncio.new_event_loop()
    loop.run_until_complete(
      self.streamNoiseModellingCmd(self.NOISEMODELLING["CMD"], feedback)
    )
    loop.close()
    
  
  # to stream the NoiseModelling script
  async def streamNoiseModellingCmd(self, cmd: str, feedback: QgsProcessingFeedback) -> None:
    proc = await asyncio.create_subprocess_shell(
      cmd,
      stdout=asyncio.subprocess.PIPE,
      stderr=asyncio.subprocess.PIPE,
      cwd = self.NOISEMODELLING["TEMP_DIR"]
    )

    while True:
      if proc.stdout.at_eof() or proc.stderr.at_eof():
        break  

      stderr_raw = await proc.stderr.readline() # for debugging
      try:
        stderr = stderr_raw.decode()
      except:
        stderr = ""
        feedback.reportError(self.tr("NoiseModelling script was not successfully executed."), fatalError=True)
        raise Exception(self.tr("NoiseModelling script was not successfully executed."))
      
      if stderr:
        feedback.pushConsoleInfo(stderr.replace("\n",""))
        
        
      prg_match = re.search(r".*[0-9]+\.[0-9]+.*%", stderr)
      if prg_match:                
        feedback.setProgress(
          int(float(re.search(r"[0-9]+\.[0-9]+", prg_match.group()).group()))
        )
  
  # import the results of the NoiseModelling as a sink
  def importNoiseModellingResultsAsSink(self, parameters: dict, context: QgsProcessingContext, attribute: str, path: str) -> None:

    dest_id = None
    if os.path.exists(path):
      layer = QgsVectorLayer(path, "layer_from_NoiseModelling")
      if layer.featureCount() > 0:
        (sink, dest_id) = self.parameterAsSink(
          parameters, attribute, context,
          layer.fields(), layer.wkbType(), QgsCoordinateReferenceSystem("EPSG:" + self.NOISEMODELLING["WPS_ARGS"]["inputSRID"])
        )
        sink.addFeatures(layer.getFeatures())
          
    return dest_id

  def name(self):
    return self.__class__.__name__

  def tr(self, string):
    return QCoreApplication.translate(self.__class__.__name__, string)
